
from . import *
from ..attribute import initializer_method

from_tiff_file  = initializer_method( ImagesInitializer.from_tiff_file  )
from_tiff_files = initializer_method( ImagesInitializer.from_tiff_files )

