
from . import *
from ..attribute import initializer_method

from_single_particle = initializer_method( TrackerInitializer.from_single_particle )
from_non_tracking    = initializer_method( TrackerInitializer.from_non_tracking    )

