
from . import *
from ..attribute import initializer_method

from_sliding_window = initializer_method( TimeInitializer.from_sliding_window )
from_sampling       = initializer_method( TimeInitializer.from_sampling       )

