



pytest tests/test_core.py tests/test_analyzer.py::TestSptDataAccess tests/test_analyzer.py::TestTesseller tests/test_analyzer.py::TestIndexer tests/test_analyzer.py::TestAssignment tests/test_analyzer.py::TestSideEffects
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
