
from . import *
from ..attribute import initializer_method

from_plugin   = initializer_method( MapperInitializer.from_plugin   )
from_callable = initializer_method( MapperInitializer.from_callable )

